#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "SDL.h"
#include "SDL_gfxPrimitives.h"

/* Adam's Bouncy Thing. */

#define WIDTH 640
#define HEIGHT 480
double efficiency = 0.7;
double friction = 0.95;
double radius = 30;
double shove = 1.5;
double gravity = 0.5;
double springlength = 10;
double springconstant = 0.1;

typedef struct {
	char *name;
	double *value;
	double step;
} twiddle;

twiddle twiddles[] = {
	{ "Efficiency", &efficiency, 0.05 },
	{ "Friction", &friction, 0.01 },
	{ "Radius", &radius, 5 },
	{ "Shove", &shove, 0.2 },
	{ "Gravity", &gravity, 0.1 },
	{ "SpringLength", &springlength, 1 },
	{ "SpringConstant", &springconstant, 0.1 }
};
#define NUMTWIDDLES (sizeof twiddles / sizeof twiddles[0])

SDL_Surface *screen;
Uint32 black;

typedef struct {
	double x;
	double y;
	double vx;
	double vy;
	double ax;
	double ay;
	double mass;
	int fixed;
} node;

node *new_node(double x, double y, double mass, int fixed) {
	node *n = malloc(sizeof *n);
	n->x = x;
	n->y = y;
	n->vx = 0;
	n->vy = 0;
	n->mass = mass;
	n->fixed = fixed;
	return n;
}

typedef struct {
	node *a;
	node *b;
	double len;
	double k;
	double force;
} spring;

spring *new_spring(node *a, node *b, double len, double k) {
	spring *s = malloc(sizeof *s);
	s->a = a;
	s->b = b;
	s->len = len;
	s->k = k;
	return s;
}

void sdldie(char *msg) {
	fprintf(stderr, "%s: %s\n", msg, SDL_GetError());
	exit(20);
}

void cls() {
	SDL_Rect clip;
	clip.x = clip.y = 0;
	clip.w = WIDTH - 1; clip.h = HEIGHT - 1;
	SDL_FillRect(screen, NULL, black);
	SDL_SetClipRect(screen, &clip);
}

void show() {
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

int main(int argc, char **argv) {
	int i;
#define NUMNODES 10
#define NUMSPRINGS 9
	node *nodes[NUMNODES];
	spring *springs[NUMSPRINGS];
	int down = 0, selected = 0;
	node *selnode = NULL;

	for (i = 0; i < 10; i++) {
		nodes[i] = new_node(100 + (10 * i), 300 + (20 * (i % 3)), 1.7 + (0.1 * i), 0);
	}
	nodes[0]->x -= 10;
	nodes[0]->mass = 3;
	nodes[0]->fixed = 1;
	nodes[9]->fixed = 1;
	nodes[9]->x = 3*WIDTH/4;
	nodes[9]->y = 1*HEIGHT/4;
	for (i = 0; i < 9; i++) {
		springs[i] = new_spring(nodes[i], nodes[i+1], 10, 0.1);
	}
	
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
		sdldie("unable to start SDL");
	atexit(SDL_Quit);

	screen = SDL_SetVideoMode(WIDTH, HEIGHT, 16, 0);
	if (screen == NULL)
		sdldie("unable to set video mode");
	black = SDL_MapRGBA(screen->format, 0, 0, 0, 0);

	SDL_WM_SetCaption("phys", "phys");
	SDL_EnableKeyRepeat(100, 20);

	while (1) {
		SDL_Event event;

		cls();

		stringRGBA(screen, 0, 0, "Adam's Bouncy Thing", 127, 255, 255, 255);
		stringRGBA(screen, 0, 10, "Q to decrement, W to increment", 255, 127, 127, 255);
		stringRGBA(screen, 0, 20, "Numbers to select variable", 255, 127, 127, 255);
		stringRGBA(screen, 0, 30, "LMB drag to shove nodes, RMB drag to move nodes", 255, 127, 127, 255);
		stringRGBA(screen, 0, 40, "MMB click to toggle node fixedness", 255, 127, 127, 255);
		for (i = 0; i < NUMTWIDDLES; i++) {
			char buf[80];
			snprintf(buf, 80, "%c%1d %-15s %lf", selected == i ? '>' : ' ', i, twiddles[i].name, *twiddles[i].value);
			stringRGBA(screen, 0, (i + 5) * 10, buf, 255, 255, 255, 255);
		}

		for (i = 0; i < NUMSPRINGS; i++) {
			spring *s = springs[i];
			s->k = springconstant;
			s->len = springlength;
		}

		for (i = 0; i < NUMNODES; i++) {
			node *n = nodes[i];
			filledCircleRGBA(screen, n->x, n->y, 2 * n->mass * n->mass, 0, n->fixed ? 100 : 0, n->fixed ? 0 : 100, 255);
		}
		for (i = 0; i < NUMSPRINGS; i++) {
			spring *s = springs[i];
			int g = 20 * abs(s->force);
			int b = g - 256;
			if (g > 255) g = 255;
			if (b < 0) b = 0;
			if (b > 255) b = 255;
			lineRGBA(screen, s->a->x, s->a->y, s->b->x, s->b->y, 255, g, b, 255);
		}
		show();

		for (i = 0; i < NUMNODES; i++) {
			node *n = nodes[i];
			n->ax = 0;
			n->ay = gravity;
		}
		for (i = 0; i < NUMSPRINGS; i++) {
			spring *s = springs[i];
			double dx = s->a->x - s->b->x;
			double dy = s->a->y - s->b->y;
			double d = sqrt((dx*dx) + (dy*dy));
			double f = (d - s->len) * s->k * efficiency;
			s->force = f;
			dx = (dx / d) * f;
			dy = (dy / d) * f;
			s->a->ax -= dx / s->a->mass;
			s->a->ay -= dy / s->a->mass;
			s->b->ax += dx / s->b->mass;
			s->b->ay += dy / s->b->mass;
		}

		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_MOUSEBUTTONDOWN: {
				int x = event.button.x, y = event.button.y;
				node *nearest = NULL;
				for (i = 0; i < NUMNODES; i++) {
					node *n = nodes[i];
					int dx = n->x - x, dy = n->y - y;
					if (sqrt((dx*dx) + (dy*dy)) < radius)
						nearest = n;
				}
				switch (event.button.button) {
				case SDL_BUTTON_LEFT:
					down = 1;
					break;
				case SDL_BUTTON_RIGHT:
					down = 2;
					selnode = nearest;
					break;
				case SDL_BUTTON_MIDDLE:
					if (nearest)
						nearest->fixed = !nearest->fixed;
					break;
				}
				break;
			}
			case SDL_MOUSEBUTTONUP:
				down = 0;
				selnode = NULL;
				break;
			case SDL_MOUSEMOTION: {
				int x = event.motion.x, y = event.motion.y;
				if (down == 2 && selnode != NULL) {
					selnode->x = x;
					selnode->y = y;
				} else if (down == 1) {
					for (i = 0; i < NUMNODES; i++) {
						node *n = nodes[i];
						int dx = n->x - x, dy = n->y - y;
						if (sqrt((dx*dx) + (dy*dy)) < radius) {
							n->ax += event.motion.xrel * shove;
							n->ay += event.motion.yrel * shove;
						}
					}
				}
				break;
			}
			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_0)
					selected = 0;
				else if (event.key.keysym.sym == SDLK_1)
					selected = 1;
				else if (event.key.keysym.sym == SDLK_2)
					selected = 2;
				else if (event.key.keysym.sym == SDLK_3)
					selected = 3;
				else if (event.key.keysym.sym == SDLK_4)
					selected = 4;
				else if (event.key.keysym.sym == SDLK_5)
					selected = 5;
				else if (event.key.keysym.sym == SDLK_6)
					selected = 6;
				else if (event.key.keysym.sym == SDLK_7)
					selected = 7;
				else if (event.key.keysym.sym == SDLK_8)
					selected = 8;
				else if (event.key.keysym.sym == SDLK_9)
					selected = 9;
				else if (event.key.keysym.sym == SDLK_q)
					*twiddles[selected].value -= twiddles[selected].step;
				else if (event.key.keysym.sym == SDLK_w)
					*twiddles[selected].value += twiddles[selected].step;
				if (selected >= NUMTWIDDLES)
					selected = 0;
				break;
			case SDL_QUIT:
				exit(0);
			}
		}

		for (i = 0; i < NUMNODES; i++) {
			node *n = nodes[i];
			if (n->fixed || n == selnode) continue;
			n->vx += n->ax;
			n->vy += n->ay;
			n->vx *= friction;
			n->vy *= friction;
			n->x += n->vx;
			n->y += n->vy;
		}

		SDL_Delay(50);
	}

	return 0;
}

